# Copyright (C) 2013 by Brian Neal.
# This file is part of purple, the PURPLE (Cipher Machine 97) simulation.
# purple is released under the MIT License (see LICENSE.txt).

"""This module contains the SteppingSwitch class and a factory function to
create the standard switches used on the PURPLE machine.

"""
import purple.data as data

# "Enum" to name the standard switches:
SIXES, TWENTIES_1, TWENTIES_2, TWENTIES_3 = range(4)


class SteppingSwitchError(Exception):
    """Exception class for all stepping switch errors"""


class SteppingSwitch:
    """This class simulates a stepping switch, the primary cryptographic element
    in the PURPLE cipher machine.

    """
    def __init__(self, dec_wiring, enc_wiring=None, init_pos=0):
        """Construct a SteppingSwitch.

        dec_wiring is the decrypt wiring table, which must be a list
        of lists (one list per input level). Each inner list is a list of
        integers representing the output contacts.

        enc_wiring is the encrypt wiring table. If None, a reciprocal table is
        built from the dec_wiring parameter.

        init_pos is the initial position of the switch; this defaults to 0.

        """
        self.dec_wiring = dec_wiring
        self.num_positions = len(dec_wiring)
        self.num_levels = len(dec_wiring[0])

        if not all(self.num_levels == len(level) for level in self.dec_wiring):
            raise SteppingSwitchError("Invalid decrypt wiring dimensions")

        self.enc_wiring = (enc_wiring if enc_wiring else
                data.build_encrypt_wiring(dec_wiring))

        if self.num_positions != len(self.enc_wiring):
            raise SteppingSwitchError("Encrypt/Decrypt positions mismatch")

        if not all(self.num_levels == len(level) for level in self.enc_wiring):
            raise SteppingSwitchError("Invalid encrypt wiring dimensions")

        self.set_pos(init_pos)

    def set_pos(self, pos):
        """Set the switch position to pos.
        Raises a SteppingSwitchError if pos is out of range.

        """
        if not (0 <= pos < self.num_positions):
            raise SteppingSwitchError("Illegal switch position")
        self.pos = pos

    def step(self):
        """Advance the stepping switch position.

        The new 0-based position of the switch is returned.

        """
        self.pos = (self.pos + 1) % self.num_positions
        return self.pos

    def decrypt(self, level):
        """This method is how to determine the output signal from the stepping
        switch for the decrypt path. The parameter 'level' is the integer input
        level for the incoming signal. The integer outgoing contact level is
        returned.

        """
        return self.dec_wiring[self.pos][level]

    def encrypt(self, level):
        """This method is how to determine the output signal from the stepping
        switch for the encrypt path. The parameter 'level' is the integer input
        level for the incoming signal. The integer outgoing contact level is
        returned.

        """
        return self.enc_wiring[self.pos][level]


def create_switch(switch_type, init_pos=0):
    """Factory function for building a SteppingSwitch of the requested
    standard type. The initial position of the switch can be specified.

    The switch_type parameter must be one of the module level constants:
        * SIXES
        * TWENTIES_1
        * TWENTIES_2
        * TWENTIES_3

    A ValueError will be raised if switch_type is an illegal value.

    """
    if switch_type not in [SIXES, TWENTIES_1, TWENTIES_2, TWENTIES_3]:
        raise ValueError("illegal switch type")

    return SteppingSwitch(dec_wiring=data.DECRYPT_DATA[switch_type],
            enc_wiring=data.ENCRYPT_DATA[switch_type],
            init_pos=init_pos)



# Copyright (C) 2013 - 2014 by Brian Neal.
# This file is part of purple, the PURPLE (Cipher Machine 97) simulation.
# purple is released under the MIT License (see LICENSE.txt).

"""This file contains the main routine for the command-line purple97 program."""

import argparse
import os
import string
import sys
import textwrap

from purple.machine import Purple97, Purple97Error
from purple.switch import SteppingSwitchError


DESC = """PURPLE cipher machine simulator"""

EPILOG = """\
Supply either -e or -d, but not both, to perform either an encrypt or decrypt. If
the -s option is not supplied, the value of the environment variable
PURPLE97_SWITCHES will be used. If the -a option is not supplied, the value of
the environment variable PURPLE97_ALPHABET will be used. Input text is supplied
either by the -t or by the -f options, but not both.
"""

DEFAULT_SWITCHES = '1-1,1,1-12'
DEFAULT_ALPHABET = Purple97.STRAIGHT_PLUGBOARD

LOWERCASE = set(string.ascii_lowercase)
DIGITS = {
    '0': 'ZERO',
    '1': 'ONE',
    '2': 'TWO',
    '3': 'THREE',
    '4': 'FOUR',
    '5': 'FIVE',
    '6': 'SIX',
    '7': 'SEVEN',
    '8': 'EIGHT',
    '9': 'NINE',
}


def filter_plaintext(source):
    """A generator to filter plaintext to ensure only valid input is fed to
    the purple cipher machine:

        * Uppercase letters A-Z are passed unchanged
        * Lowercase letters a-z are converted to uppercase
        * Digits are converted to words; e.g. 1 => ONE
        * All other input is ignored (filtered out)

    source can be anything that yields a line of text at a time.

    """
    for line in source:
        for c in line:
            if c in Purple97.VALID_KEYS:
                yield c
            elif c in LOWERCASE:
                yield c.upper()
            else:
                digit_name = DIGITS.get(c)
                if digit_name:
                    for d in digit_name:
                        yield d


def filter_whitespace(source):
    """A generator to filter out whitespace from text read from
    a file-like source.

    """
    for line in source:
        for c in line:
            if not c.isspace():
                yield c


def group_text(text, n=5):
    """Groups the given text into n-letter groups separated by spaces."""

    return ' '.join(text[i:i+n] for i in range(0, len(text), n))


def main(argv=None):
    """Entry point for the command-line purple97 simulation."""

    parser = argparse.ArgumentParser(description=DESC, epilog=EPILOG)
    parser.add_argument('-e', '--encrypt', action='store_true',
        help='perform an encrypt operation')
    parser.add_argument('-d', '--decrypt', action='store_true',
        help='perform a decrypt operation')
    parser.add_argument('-f', '--filter', action='store_true',
        help='filter plaintext and provide useful substitutions')
    parser.add_argument('-s', '--switches',
        help='switch settings, e.g. 9-1,24,6-23')
    parser.add_argument('-a', '--alphabet',
        help='plugboard wiring string, 26-letters; e.g. AEIOUYBCDFGHJKLMNPQRSTVWXZ')
    parser.add_argument('-t', '--text',
        help='input text to encrypt/decrypt')
    parser.add_argument('-i', '--input', metavar='FILE',
        help='file to read input text from, - for stdin')
    parser.add_argument('-g', '--group', type=int, default=5, metavar='N',
        help=('if non-zero, group output in %(metavar)s-letter groups '
              '[default: %(default)s]'))
    parser.add_argument('-w', '--width', type=int, default=70, metavar='N',
        help=("wrap output text to %(metavar)s letters; "
              "a value of 0 means do not wrap "
              "[default: %(default)s]"))

    args = parser.parse_args(args=argv)

    if not args.encrypt and not args.decrypt:
        parser.print_help()
        parser.exit(1)
    if args.encrypt and args.decrypt:
        parser.error("Please supply either -e or -d, not both")
    if args.text and args.input:
        parser.error("Please supply either -t or -i, not both")
    if args.decrypt and args.filter:
        parser.error("The -f option only works with -e (encrypt)")

    # Get key settings
    if args.switches:
        switches = args.switches
    else:
        switches = os.environ.get('PURPLE97_SWITCHES', DEFAULT_SWITCHES)

    if args.alphabet:
        alphabet = args.alphabet
    else:
        alphabet = os.environ.get('PURPLE97_ALPHABET', DEFAULT_ALPHABET)

    if args.group < 0:
        parser.error("The --group option must be 0 or greater")
    if args.width < 0:
        parser.error("The --width option must be 0 or greater")

    # Create purple cipher machine object
    try:
        purple = Purple97.from_key_sheet(switches, alphabet)
    except (Purple97Error, SteppingSwitchError) as ex:
        parser.error(str(ex))


    if args.input:
        try:
            fp = sys.stdin if args.input == '-' else open(args.input, 'r')
        except IOError as ex:
            raise SystemExit(str(ex))

    if args.encrypt:
        action = purple.encrypt

        source = [args.text] if args.text else fp
        source = (filter_plaintext(source) if args.filter else
                                    filter_whitespace(source))
    else:
        action = purple.decrypt
        source = [args.text] if args.text else fp
        source = filter_whitespace(source)

    try:
        result = action(source)
    except (Purple97Error, SteppingSwitchError) as ex:
        parser.error(str(ex))

    if args.group != 0:
        result = group_text(result, args.group)

    if args.width != 0:
        # Never break groups when grouping
        break_long_words = args.group == 0
        result = textwrap.fill(result, width=args.width, expand_tabs=False,
                replace_whitespace=False, drop_whitespace=False,
                break_long_words=break_long_words)

    print(result)


if __name__ == '__main__':
    main()



# Copyright (C) 2013 - 2014 by Brian Neal.
# This file is part of purple, the PURPLE (Cipher Machine 97) simulation.
# purple is released under the MIT License (see LICENSE.txt).

"""This module contains the Purple97 class, the top level class in the PURPLE
simulation.

The reference for this simulation is the paper "PURPLE Revealed: Simulation and
Computer-aided Cryptanalysis of Angooki Taipu B", by Wes Freeman, Geoff
Sullivan, and Frode Weierud. This paper was published in Cryptologia, Volume 27,
Issue 1, January, 2003, pp. 1-43.

The paper is also available at:
http://cryptocellar.web.cern.ch/cryptocellar/pubs/PurpleRevealed.pdf

"""
from collections import Counter
import string

import purple.switch as switch


class Purple97Error(Exception):
    """Exception class for all Purple97 errors"""


class Purple97:
    """This class simulates the top-level behavior of the PURPLE cipher
    machine.

    """
    VALID_KEYS = set(string.ascii_uppercase)
    STRAIGHT_PLUGBOARD = 'AEIOUYBCDFGHJKLMNPQRSTVWXZ'

    def __init__(self, switches_pos=None, fast_switch=1, middle_switch=2,
            alphabet=None):
        """Build a PURPLE (Cipher Machine 97) instance. Initial settings can be
        optionally supplied.

        switches_pos: If not None, must be a 4-element list or tuple of integer
        starting switch positions. Each position must be in the range of 0-24,
        inclusive. If None, a list of all 0's is assumed. The first element in
        the list is the starting position for the sixes switch. The second
        through fourth elements are the starting positions for the three
        twenties switches, 1 through 3.

        fast_switch: this integer parameter names which twenties switch (1-3) is
        assuming the role of the fast switch.

        middle_switch: this integer parameter names which twenties switch (1-3)
        has been designated as the middle switch.

        Passing in the same value for both the fast and medium switches will
        raise a Purple97Error exception. The slow switch is assumed to be the
        remaining twenties switch that was not named.

        alphabet: this parameter must be a 26-letter sequence that represents
        the daily alphabet setting. It describes how the typewriters are wired
        to the plugboard. The first six characters are the mapping for the sixes
        switch, and the remaining 20 are for the input wiring for the first
        stage of the twenties switches. If None is supplied, a straight through
        mapping is assumed; i.e. STRAIGHT_PLUGBOARD.

        The alphabet parameter will accept either upper or lowercase letters.
        All 26 distinct letters must be present or else a Purple97Error
        exception will be raised.

        """
        # If no switch positions are supplied, default to all 0's
        if switches_pos is None:
            switches_pos = (0, 0, 0, 0)

        # Validate switch positions
        try:
            n = len(switches_pos)
        except TypeError:
            raise Purple97Error("switches_pos must be a sequence")
        if n != 4:
            raise Purple97Error("switches_pos must have length of 4")

        # Create switches with correct starting positions
        self.sixes = switch.create_switch(switch.SIXES, switches_pos[0])
        self.twenties = [
            switch.create_switch(switch.TWENTIES_1, switches_pos[1]),
            switch.create_switch(switch.TWENTIES_2, switches_pos[2]),
            switch.create_switch(switch.TWENTIES_3, switches_pos[3]),
        ]

        # Validate fast & middle switch parameters
        if not (1 <= fast_switch <= 3):
            raise Purple97Error("fast_switch out of range (1-3)")
        if not (1 <= middle_switch <= 3):
            raise Purple97Error("middle_switch out of range (1-3)")
        if fast_switch == middle_switch:
            raise Purple97Error("fast & middle switches cannot be the same")

        # Store references to the fast, middle, and slow switches
        self.fast_switch = self.twenties[fast_switch - 1]
        self.middle_switch = self.twenties[middle_switch - 1]

        # Pick the remaining switch as the slow switch
        switches = [1, 2, 3]
        switches.remove(fast_switch)
        switches.remove(middle_switch)
        self.slow_switch = self.twenties[switches[0] - 1]

        # Validate the alphabet
        if alphabet is None:
            alphabet = self.STRAIGHT_PLUGBOARD

        if len(alphabet) != 26:
            raise Purple97Error("invalid alphabet length")
        alphabet = alphabet.upper()

        # Count valid letters
        ctr = Counter(string.ascii_uppercase)
        for c in alphabet:
            if c in ctr:
                ctr[c] += 1

        # At this point if alphabet is legal, all keys in ctr must have a value
        # of 2. If any are 1, then alphabet was missing some letters. If any are
        # greater than 2, we had duplicate letters.

        if not all(v == 2 for v in ctr.values()):
            raise Purple97Error("invalid alphabet")

        self.alphabet = alphabet
        self.plugboard = {c : n for n, c in enumerate(alphabet)}

    @classmethod
    def from_key_sheet(cls, switches, alphabet=None):
        """This class method allows one to construct a Purple97 using
        a shorthand notation used by US codebreakers.

        switches: must be a string of the form 'a-b,c,d-ef' where
            a - starting position of the sixes switch (1-25)
            b - starting position of the twenties switch #1 (1-25)
            c - starting position of the twenties switch #2 (1-25)
            d - starting position of the twenties switch #3 (1-25)
            e - which switch is the fast switch (1-3)
            f - which switch is the middle switch (1-3)

        Example: '9-1,24,6-23'

        Note that the starting positions here are 1-based since that is the
        notation the US codebreakers seemed to have used.

        alphabet: the daily alphabet, same format as in the __init__ function

        """
        try:
            sixes, twenties, speed = switches.split('-')
        except ValueError:
            raise Purple97Error('invalid switches string (-)')

        twenties = twenties.split(',')
        if len(twenties) != 3:
            raise Purple97Error('invalid twenties position')

        try:
            switches_pos = [int(s) - 1 for s in [sixes] + twenties]
        except ValueError:
            raise Purple97Error('switch positions must be numeric')

        if len(speed) != 2:
            raise Purple97Error('invalid switch speed settings')

        try:
            fast_switch, middle_switch = int(speed[0]), int(speed[1])
        except ValueError:
            raise Purple97Error('switch speed settings must be numeric')

        return cls(switches_pos, fast_switch, middle_switch, alphabet)

    def decrypt(self, ciphertext):
        """Decrypts the given ciphertext message and returns the plaintext
        output.

        ciphertext must contain only the letters A-Z or '-' or else a
        Purple97Error exception is raised. A '-' is used to indicate a garble.
        When a '-' is encountered, a '-' is added to the plaintext output and
        the machine is stepped.

        """
        plaintext = []
        for i, c in enumerate(ciphertext):

            # Process a garble:
            if c == '-':
                plaintext.append('-')
                self.step()
                continue

            if c not in self.VALID_KEYS:
                raise Purple97Error("invalid input '{}' to decrypt".format(c))

            n = self.plugboard[c]

            if n < 6:
                # This input goes to the sixes switch
                x = self.sixes.decrypt(n)
            else:
                # This input goes to the chain of twenties switches
                n -= 6
                x = self.twenties[0].decrypt(self.twenties[1].decrypt(
                        self.twenties[2].decrypt(n)))
                x += 6

            plaintext.append(self.alphabet[x])

            # Now step the switches.
            self.step()

        return ''.join(plaintext)

    def encrypt(self, plaintext):
        """Encrypts the given plaintext message and returns the ciphertext
        output.

        plaintext must contain only the letters A-Z or else a Purple97Error
        exception is raised.

        """
        ciphertext = []
        for c in plaintext:
            if c not in self.VALID_KEYS:
                raise Purple97Error("invalid input '{}' to encrypt".format(c))

            n = self.plugboard[c]
            if n < 6:
                # This input goes to the sixes switch
                x = self.sixes.encrypt(n)
            else:
                # This input goes to the chain of twenties switches in reverse
                # order compared to decrypt.
                n -= 6
                x = self.twenties[2].encrypt(self.twenties[1].encrypt(
                        self.twenties[0].encrypt(n)))
                x += 6

            ciphertext.append(self.alphabet[x])

            # Now step the switches.
            self.step()

        return ''.join(ciphertext)

    def step(self):
        """Step the stepping switches."""
        # First read the sixes and middle switch
        # positions before stepping anything. Use these latched values in
        # the decision processes for stepping a twenties. This is crucial!
        sixes_pos = self.sixes.pos
        middle_pos = self.middle_switch.pos

        # Now we can step the sixes. It unconditionally steps after every
        # letter is processed.
        self.sixes.step()

        # Only 1 twenties switch steps at a time.
        #
        # Normally the fast switch advances every letter.
        #
        # However if the sixes is at the last position (24), the middle
        # switch steps instead.
        #
        # But if the sixes is at the second to last position (23) and the middle
        # switch is at the last position (24), the slow switch will step.  The
        # middle switch will step on the next letter in this case.

        if sixes_pos == 23 and middle_pos == 24:
            self.slow_switch.step()
        elif sixes_pos == 24:
            self.middle_switch.step()
        else:
            self.fast_switch.step()



# Copyright (C) 2013 - 2014 by Brian Neal.
# This file is part of purple, the PURPLE (Cipher Machine 97) simulation.
# purple is released under the MIT License (see LICENSE.txt).

import string
import unittest

from purple.machine import Purple97, Purple97Error
from purple.switch import SteppingSwitchError


# This is part 1 of the famous 14 part message. The ciphertext and plaintext are
# laid on top of each other, every other line, with blank lines inserted for
# clarity. Garbles are indicated by '-' characters.
#
part1 = """
ZTXODNWKCCMAVNZXYWEETUQTCIMNVEUVIWBLUAXRRTLVA
FOVTATAKIDASINIMUIMINOMOXIWOIRUBESIFYXXFCKZZR

RGNTPCNOIUPJLCIVRTPJKAUHVMUDTHKTXYZELQTVWGBUHFAWSH
DXOOVBTNFYXFAEMEMORANDUMFIOFOVOOMOJIBAKARIFYXRAICC

ULBFBHEXMYHFLOWD-KWHKKNXEBVPYHHGHEKXIOHQHUHWIKYJYH
YLFCBBCFCTHEGOVE-NMENTOFJAPANLFLPROMPTEDBYAGENUINE

PPFEALNNAKIBOOZNFRLQCFLJTTSSDDOIOCVT-ZCKQTSHXTIJCN
DESIRETOCOMETOANAMICABLEUNDERSTANDIN-WITHTHEGOVERN

WXOKUFNQR-TAOIHWTATWVHOTGCGAKVANKZANMUIN
MENTOFTHE-NITEDSTATESINORDERTHATTHETWOCO

YOYJFSRDKKSEQBWKIOORJAUWKXQGUWPDUDZNDRMDHVHYPNIZXB
UNTRIESBYTHEIRJOINTEFFORTSMAYSECURETHEPEACEOFTHEPA

GICXRMAWMFTIUDBXIENLONOQVQKYCOTVSHVNZZQPDLMXVNRUUN
CIFICAREAANDTHEREBYCONTRIBUTETOWARDTHEREALIZATIONO

QFTCDFECZDFGMXEHHWYONHYNJDOVJUNCSUVKKEIWOLKRBUUSOZ
FWORLDPEACELFLHASCONTINUEDNEGOTIATIONSWITHTHEUTMOS

UIGNISMWUOSBOBLJXERZJEQYQMTFTXBJNCMJKVRKOTSOPBOYMK
TSINCERITYSINCEAPRILLASTWITHTHEGOVERNMENTOFTHEUNIT

IRETINCPSQJAWVHUFKRMAMXNZUIFNOPUEMHGLOEJHZOOKHHEED
EDSTATESREGARDINGTHEADJUSTMENTANDADVANCEMENTOFJAPA

NIHXFXFXGPDZBSKAZABYEKYEPNIYSHVKFRFPVCJTPTOYCNEIQB
NESEVVFAMERICANRELATIONSANDTHESTABILIZATIONOFTHEPA

FEXMERMIZLGDRXZORLZFSQYPZFATZCHUGRNHWDDTAIHYOOCOOD
CIFICAREACFCCCFTHEJAPANESEQOVERNMENXHASTHEHONORTOS

UZYIWJROOJUMUIHRBEJFONAXGNCKAOARDIHCDZKIXPR--DIMUW
TATEFRANKLYITSVIEWSCONCERNINGTHECLAIMSTHEAM--VCANG

OMHLTJSOUXPFKGEPWJOMTUVKMWRKTACUPIGAFEDFVRKXFXLFGU
OVERNMENTHASUERSISTENTLYMAINTAINEDASWELLASTHEMEASU

RDETJIYOLKBHZKXOJDDOVRHMMUQBFOWRODMRMUWNAYKYPISDLH
RESTHEUNITEDSTATESANDGREATBRITAINHAVETAKENTOWARDJA

ECKINLJORKWNWXADAJOLONOEVMUQDFIDSPEBBPWROFBOPAZJEU
PANDURINGTHKSEEIGHTMONTHSCYCCCFLFCDDCFCITISTHEIMMU

USBHGIORCSUUQKIIEHPCTJRWSOGLETZLOUKKEOJOSMKJBWUCDD
TABLXPOLWCYOFTHEJAPANESEGOVERNMENTTOINSURETHESTABI

CPYUUWCSSKWWVLIUPKYXGKQOKAZTEZFHGVPJFEWEUBKLIZLWKK
LITYOFEASTASIAANDTOPROMOTEWORLZPEACELFLANDTHEREBYT

OBXLEPQPDATWUSUUPKYRHNWDZXXGTWDDNSHDCBCJXAOOEEPUBP
OEIABLEALLNATIONSTOFINDEACHITSPROPERPLACEINTHEWORL

WFRBQSFXSEZJJYAANMG-WLYMGWAQDGIVNOHKOUTIXYFOKNGGBF
DCFCCCFEVERSINCETHE-HINAAFFAIRBROKEOUTOWINGTOTHEFA

GANPWTUYLBEFFKUFLEXOIUUANVMMJEQUSFHFDOHQLAKWTBYYYL
ILUREONTHEPARTOFCHINATOCOMPREHENLJAPANVCFSTRUEYNTE

NTLYTSXCGKCEEWQRYAVGRKXIANPXNOFVXGKJFAVKLTHOCXCIVK
NTIONSLFLTWEJAPANESEGOVERNMENTHASSTRIVENFORTHEREST

OLXTJTUNCLQCICRUIIWQDDMOTPRVTJKKSKFHXFKMDIKIZWROGZ
ORATIONOFPEACEANDIMHASCONSISTENTLYEXERTEDITSBESTEF

JYMTMNOVMFJ-OKTEIVMYANOHNNYPDLEXCFRRNEBLMNYEBGNHCZ
FORTSTOPREV-NTTHEEXTENTIONOFWARVVFLIKEVISTURBANCES

ZCFNWGGRHRIUUTTILKLODUYZKQOZMMNHASXHLPVTNGHQDAJIUG
CFCNSIASALSOTOTHATENDTNATINSEPTEMBERLASTYEARJAPANC

OOSZ-----ZRTGWFBLKI--------YBDABJ-----WYOEANV---OM
ONCL-----HETRIPAITI--------THGERM-----DYTALYC---OV
"""

pt1_lines = part1.split()
PT1_CT = ''.join(pt1_lines[0::2])
PT1_PT = ''.join(pt1_lines[1::2])


class Purple97TestCase(unittest.TestCase):

    def test_construction(self):

        Purple97()
        Purple97([0, 1, 2, 3])
        Purple97([0, 1, 2, 3], 1)
        Purple97([0, 1, 2, 3], 2, 1)
        Purple97((0, 1, 2, 3), 2, 1, string.ascii_uppercase)
        Purple97((0, 1, 2, 3), 2, 1, string.ascii_lowercase)
        Purple97(alphabet=string.ascii_lowercase)
        Purple97(fast_switch=3, middle_switch=1)

    def test_construct_bad_positions(self):

        self.assertRaises(Purple97Error, Purple97, [])
        self.assertRaises(Purple97Error, Purple97, [1])
        self.assertRaises(Purple97Error, Purple97, [1, 1, 1, 1, 1])
        self.assertRaises(SteppingSwitchError, Purple97, [1, 1, 1, 100])

    def test_construct_bad_switches(self):

        self.assertRaises(Purple97Error, Purple97, fast_switch=0)
        self.assertRaises(Purple97Error, Purple97, fast_switch=4)
        self.assertRaises(Purple97Error, Purple97, fast_switch=-1)

        self.assertRaises(Purple97Error, Purple97, middle_switch=0)
        self.assertRaises(Purple97Error, Purple97, middle_switch=4)
        self.assertRaises(Purple97Error, Purple97, middle_switch=-1)

        self.assertRaises(Purple97Error, Purple97, fast_switch=2)
        self.assertRaises(Purple97Error, Purple97, fast_switch=1, middle_switch=1)
        self.assertRaises(Purple97Error, Purple97, fast_switch=2, middle_switch=2)
        self.assertRaises(Purple97Error, Purple97, fast_switch=3, middle_switch=3)

        self.assertRaises(Purple97Error, Purple97, fast_switch=0, middle_switch=1)
        self.assertRaises(Purple97Error, Purple97, fast_switch=0, middle_switch=0)
        self.assertRaises(Purple97Error, Purple97, fast_switch=0, middle_switch=4)

    def test_construct_bad_alphabet(self):

        alpha = ''
        self.assertRaises(Purple97Error, Purple97, alphabet=alpha)
        alpha = '1'
        self.assertRaises(Purple97Error, Purple97, alphabet=alpha)
        alpha = '!' * 26
        self.assertRaises(Purple97Error, Purple97, alphabet=alpha)
        alpha = string.ascii_uppercase[3:]
        self.assertRaises(Purple97Error, Purple97, alphabet=alpha)
        alpha = string.ascii_uppercase + string.ascii_uppercase[4:10]
        self.assertRaises(Purple97Error, Purple97, alphabet=alpha)
        alpha = string.ascii_uppercase[:13] + string.ascii_uppercase[:13]
        self.assertRaises(Purple97Error, Purple97, alphabet=alpha)
        alpha = 'M' * 26
        self.assertRaises(Purple97Error, Purple97, alphabet=alpha)

    def test_from_key_sheet(self):

        Purple97.from_key_sheet('9-1,2,3-23')
        Purple97.from_key_sheet('1-1,1,1-13')
        Purple97.from_key_sheet('25-25,25,25-31')
        Purple97.from_key_sheet('5-20,7,18-21', alphabet=string.ascii_uppercase)

    def test_bad_from_key_sheet(self):

        self.assertRaises(SteppingSwitchError, Purple97.from_key_sheet, '0-1,2,3-13')
        self.assertRaises(SteppingSwitchError, Purple97.from_key_sheet, '26-1,2,3-13')
        self.assertRaises(SteppingSwitchError, Purple97.from_key_sheet, '1-1,0,3-13')
        self.assertRaises(SteppingSwitchError, Purple97.from_key_sheet, '1-1,2,26-13')
        self.assertRaises(SteppingSwitchError, Purple97.from_key_sheet, '1-1,2,26-03')
        self.assertRaises(SteppingSwitchError, Purple97.from_key_sheet, '1-1,2,26-00')
        self.assertRaises(SteppingSwitchError, Purple97.from_key_sheet, '1-1,2,26-14')

        self.assertRaises(Purple97Error, Purple97.from_key_sheet, 'bad string')
        self.assertRaises(Purple97Error, Purple97.from_key_sheet, '1-2-1,2,26-14')
        self.assertRaises(Purple97Error, Purple97.from_key_sheet, 'a-9,2,20-13')
        self.assertRaises(Purple97Error, Purple97.from_key_sheet, '1-a,2,20-13')
        self.assertRaises(Purple97Error, Purple97.from_key_sheet, '1-9,a,20-13')
        self.assertRaises(Purple97Error, Purple97.from_key_sheet, '1-9,2,a-13')
        self.assertRaises(Purple97Error, Purple97.from_key_sheet, '1-9,2,20-a3')
        self.assertRaises(Purple97Error, Purple97.from_key_sheet, '1-9,2,20-1a')
        self.assertRaises(Purple97Error, Purple97.from_key_sheet, '1-9,2,20-123')

    def test_decrypt_part_1_message(self):

        ciphertext = PT1_CT
        plaintext = PT1_PT

        self.assertEqual(len(ciphertext), len(plaintext))

        # Decrypt
        purple = Purple97.from_key_sheet(
                switches='9-1,24,6-23',
                alphabet='NOKTYUXEQLHBRMPDICJASVWGZF')

        actual = purple.decrypt(ciphertext)

        mismatches = []
        for n, (a, b) in enumerate(zip(plaintext, actual)):
            if a != b:
                mismatches.append((n, a, b))

        msg = None
        if mismatches:
            msg = "There are {} mismatches: {}".format(len(mismatches),
                    mismatches)

        self.assertTrue(len(mismatches) == 0, msg)

    def test_encrypt_part_1_message(self):

        ciphertext = PT1_CT
        plaintext = PT1_PT

        # Use 'X' in place of the garbles
        input_text = plaintext.replace('-', 'X')

        self.assertEqual(len(ciphertext), len(plaintext))

        # Encrypt
        purple = Purple97.from_key_sheet(
                switches='9-1,24,6-23',
                alphabet='NOKTYUXEQLHBRMPDICJASVWGZF')

        actual = purple.encrypt(input_text)

        mismatches = []
        for n, (a, b) in enumerate(zip(ciphertext, actual)):
            if a != b and a != '-':
                mismatches.append((n, a, b))

        msg = None
        if mismatches:
            msg = "There are {} mismatches: {}".format(len(mismatches),
                    mismatches)

        self.assertTrue(len(mismatches) == 0, msg)



# Copyright (C) 2013 by Brian Neal.
# This file is part of purple, the PURPLE (Cipher Machine 97) simulation.
# purple is released under the MIT License (see LICENSE.txt).

import unittest

from purple.switch import SteppingSwitch, SteppingSwitchError, create_switch
import purple.switch


class SwitchTestCase(unittest.TestCase):

    def test_construction(self):

        bad_wiring = [
            [1, 2, 0],
            [2, 1],
            [0, 1, 2]
        ]
        good_wiring = [
            [1, 2, 0],
            [2, 1, 0],
            [0, 1, 2]
        ]
        self.assertRaises(SteppingSwitchError, SteppingSwitch, bad_wiring)
        self.assertRaises(SteppingSwitchError, SteppingSwitch, bad_wiring,
                init_pos=0)
        self.assertRaises(SteppingSwitchError, SteppingSwitch, good_wiring,
                init_pos=-1)
        self.assertRaises(SteppingSwitchError, SteppingSwitch, good_wiring,
                init_pos=3)
        self.assertRaises(SteppingSwitchError, SteppingSwitch, bad_wiring,
                init_pos=3)

    def test_wiring_table_mismatch_in_dimensions(self):

        wiring_1 = [
            [1, 2, 0],
            [2, 1, 0],
            [0, 1, 2],
        ]
        wiring_2 = [
            [1, 2, 3, 0],
            [1, 2, 3, 0],
            [1, 2, 3, 0],
        ]
        wiring_3 = [
            [1, 2, 3, 0],
            [1, 2, 3, 0],
            [1, 2, 3, 0],
            [1, 2, 3, 0],
        ]
        bad_wiring = [
            [1, 2, 0],
            [2, 1],
            [0, 1, 2]
        ]
        self.assertRaises(SteppingSwitchError, SteppingSwitch, wiring_1,
                wiring_2)
        self.assertRaises(SteppingSwitchError, SteppingSwitch, wiring_2,
                wiring_1)
        self.assertRaises(SteppingSwitchError, SteppingSwitch, wiring_1,
                bad_wiring)
        self.assertRaises(SteppingSwitchError, SteppingSwitch, wiring_2,
                wiring_3)
        self.assertRaises(SteppingSwitchError, SteppingSwitch, wiring_3,
                wiring_2)

    def test_bad_set_pos(self):

        switch = create_switch(purple.switch.SIXES)
        self.assertRaises(SteppingSwitchError, switch.set_pos, -1)
        self.assertRaises(SteppingSwitchError, switch.set_pos, 25)
        self.assertRaises(SteppingSwitchError, switch.set_pos, 26)

    def test_set_pos(self):

        switch = create_switch(purple.switch.SIXES)
        for pos in range(25):
            switch.set_pos(pos)

    def test_step(self):

        switch = create_switch(purple.switch.SIXES)
        for n in range(25 * 2 + 1):
            pos = n % 25
            self.assertEqual(switch.pos, pos)
            new_pos = switch.step()
            self.assertEqual((pos + 1) % 25, new_pos)

    def test_decrypt(self):

        switch = create_switch(purple.switch.SIXES)
        switch.set_pos(0)
        self.assertEqual(switch.decrypt(0), 1)
        self.assertEqual(switch.decrypt(1), 0)
        self.assertEqual(switch.decrypt(2), 2)
        self.assertEqual(switch.decrypt(3), 4)
        self.assertEqual(switch.decrypt(4), 3)
        self.assertEqual(switch.decrypt(5), 5)

        switch.step()
        self.assertEqual(switch.decrypt(0), 5)
        self.assertEqual(switch.decrypt(1), 2)
        self.assertEqual(switch.decrypt(2), 4)
        self.assertEqual(switch.decrypt(3), 1)
        self.assertEqual(switch.decrypt(4), 0)
        self.assertEqual(switch.decrypt(5), 3)

        switch.set_pos(24)
        self.assertEqual(switch.decrypt(0), 4)
        self.assertEqual(switch.decrypt(1), 1)
        self.assertEqual(switch.decrypt(2), 3)
        self.assertEqual(switch.decrypt(3), 2)
        self.assertEqual(switch.decrypt(4), 5)
        self.assertEqual(switch.decrypt(5), 0)

    def test_encrypt(self):

        switch = create_switch(purple.switch.SIXES)
        switch.set_pos(0)
        self.assertEqual(switch.encrypt(0), 1)
        self.assertEqual(switch.encrypt(1), 0)
        self.assertEqual(switch.encrypt(2), 2)
        self.assertEqual(switch.encrypt(3), 4)
        self.assertEqual(switch.encrypt(4), 3)
        self.assertEqual(switch.encrypt(5), 5)

        switch.step()
        self.assertEqual(switch.encrypt(0), 4)
        self.assertEqual(switch.encrypt(1), 3)
        self.assertEqual(switch.encrypt(2), 1)
        self.assertEqual(switch.encrypt(3), 5)
        self.assertEqual(switch.encrypt(4), 2)
        self.assertEqual(switch.encrypt(5), 0)

        switch.set_pos(24)
        self.assertEqual(switch.encrypt(0), 5)
        self.assertEqual(switch.encrypt(1), 1)
        self.assertEqual(switch.encrypt(2), 3)
        self.assertEqual(switch.encrypt(3), 2)
        self.assertEqual(switch.encrypt(4), 0)
        self.assertEqual(switch.encrypt(5), 4)

    def test_bad_create_switch(self):

        self.assertRaises(ValueError, create_switch, -1)
        self.assertRaises(ValueError, create_switch, 4)
        self.assertRaises(ValueError, create_switch, 99)
